import argparse
from multiprocessing import Process
from pathlib import Path

import requests
from livereload import Server

_DRAWF_SOURCE_PREFIX = "drawfsource/js/fastled/src/"


def _run_flask_server(
    fastled_js: Path,
    port: int,
    compile_server_port: int,
    certfile: Path | None = None,
    keyfile: Path | None = None,
) -> None:
    """Run Flask server with live reload in a subprocess

    Args:
        fastled_js: Path to the fastled_js directory
        port: Port to run the server on
        certfile: Path to the SSL certificate file
        keyfile: Path to the SSL key file
    """
    try:
        from flask import Flask, Response, send_from_directory

        app = Flask(__name__)

        # Must be a full path or flask will fail to find the file.
        fastled_js = fastled_js.resolve()

        print(f"Compile server port is at {compile_server_port}")

        @app.route("/")
        def serve_index():
            return send_from_directory(fastled_js, "index.html")

        @app.route("/sourcefiles/<path:path>")
        def serve_source_files(path):
            """Proxy requests to /sourcefiles/* to the compile server"""
            from flask import request

            # Forward the request to the compile server
            target_url = f"http://localhost:{compile_server_port}/sourcefiles/{path}"

            # Forward the request with the same method, headers, and body
            resp = requests.request(
                method=request.method,
                url=target_url,
                headers={key: value for key, value in request.headers if key != "Host"},
                data=request.get_data(),
                cookies=request.cookies,
                allow_redirects=True,
                stream=False,
            )

            # Create a Flask Response object from the requests response
            response = Response(
                resp.raw.read(), status=resp.status_code, headers=dict(resp.headers)
            )

            return response

        def handle_dwarfsource(path: str) -> Response:
            """Handle requests to /drawfsource/js/fastled/src/"""
            from flask import request

            print("\n##################################")
            print(f"# Serving source file /drawfsource/ {path}")
            print("##################################\n")

            if not path.startswith(_DRAWF_SOURCE_PREFIX):
                # unexpected
                print(f"Unexpected path: {path}")
                return Response("Malformed path", status=400)

            path = path.replace("drawfsource/js/fastled/src/", "")

            # Forward the request to the compile server
            target_url = f"http://localhost:{compile_server_port}/sourcefiles/{path}"

            # Forward the request with the same method, headers, and body
            resp = requests.request(
                method=request.method,
                url=target_url,
                headers={key: value for key, value in request.headers if key != "Host"},
                data=request.get_data(),
                cookies=request.cookies,
                allow_redirects=True,
                stream=False,
            )

            # Create a Flask Response object from the requests response
            response = Response(
                resp.raw.read(), status=resp.status_code, headers=dict(resp.headers)
            )

            return response

        def handle_sourcefile(path: str) -> Response:
            """Handle requests to /sourcefiles/*"""
            from flask import Response, request

            print("\n##################################")
            print(f"# Serving source file /sourcefiles/ {path}")
            print("##################################\n")

            # Forward the request to the compile server
            target_url = f"http://localhost:{compile_server_port}/sourcefiles/{path}"

            # Forward the request with the same method, headers, and body
            resp = requests.request(
                method=request.method,
                url=target_url,
                headers={key: value for key, value in request.headers if key != "Host"},
                data=request.get_data(),
                cookies=request.cookies,
                allow_redirects=True,
                stream=False,
            )

            # Create a Flask Response object from the requests response
            response = Response(
                resp.raw.read(), status=resp.status_code, headers=dict(resp.headers)
            )

            return response

        def handle_local_file_fetch(path: str) -> Response:

            print("\n##################################")
            print(f"# Servering generic file {path}")
            print("##################################\n")

            response = send_from_directory(fastled_js, path)
            # Some servers don't set the Content-Type header for a bunch of files.
            if path.endswith(".js"):
                response.headers["Content-Type"] = "application/javascript"
            if path.endswith(".css"):
                response.headers["Content-Type"] = "text/css"
            if path.endswith(".wasm"):
                response.headers["Content-Type"] = "application/wasm"
            if path.endswith(".json"):
                response.headers["Content-Type"] = "application/json"
            if path.endswith(".png"):
                response.headers["Content-Type"] = "image/png"
            if path.endswith(".jpg"):
                response.headers["Content-Type"] = "image/jpeg"
            if path.endswith(".jpeg"):
                response.headers["Content-Type"] = "image/jpeg"
            if path.endswith(".gif"):
                response.headers["Content-Type"] = "image/gif"
            if path.endswith(".svg"):
                response.headers["Content-Type"] = "image/svg+xml"
            if path.endswith(".ico"):
                response.headers["Content-Type"] = "image/x-icon"
            if path.endswith(".html"):
                response.headers["Content-Type"] = "text/html"

            # now also add headers to force no caching
            response.headers["Cache-Control"] = "no-cache, no-store, must-revalidate"
            response.headers["Pragma"] = "no-cache"
            response.headers["Expires"] = "0"
            return response

        @app.route("/<path:path>")
        def serve_files(path: str):

            if path.startswith("drawfsource/"):
                return handle_dwarfsource(path)
            elif path.startswith("sourcefiles/"):
                return handle_sourcefile(path)
            else:
                return handle_local_file_fetch(path)

        server = Server(app.wsgi_app)
        # Watch index.html for changes
        server.watch(str(fastled_js / "index.html"))
        # server.watch(str(fastled_js / "index.js"))
        # server.watch(str(fastled_js / "index.css"))
        # Start the server
        server.serve(port=port, debug=True)
    except KeyboardInterrupt:
        import _thread

        _thread.interrupt_main()
    except Exception as e:
        print(f"Failed to run Flask server: {e}")
        import _thread

        _thread.interrupt_main()


def run_flask_in_thread(
    port: int,
    cwd: Path,
    compile_server_port: int,
    certfile: Path | None = None,
    keyfile: Path | None = None,
) -> None:
    """Run the Flask server."""
    try:
        _run_flask_server(cwd, port, compile_server_port, certfile, keyfile)
        import warnings

        warnings.warn("Flask server has stopped")
    except KeyboardInterrupt:
        import _thread

        _thread.interrupt_main()
        pass


def parse_args() -> argparse.Namespace:
    """Parse the command line arguments."""
    parser = argparse.ArgumentParser(
        description="Open a browser to the fastled_js directory"
    )
    parser.add_argument(
        "fastled_js", type=Path, help="Path to the fastled_js directory"
    )
    parser.add_argument(
        "--port",
        "-p",
        type=int,
        required=True,
        help="Port to run the server on (default: %(default)s)",
    )
    parser.add_argument(
        "--certfile",
        type=Path,
        help="Path to the SSL certificate file for HTTPS",
    )
    parser.add_argument(
        "--keyfile",
        type=Path,
        help="Path to the SSL key file for HTTPS",
    )
    return parser.parse_args()


def run_flask_server_process(
    port: int,
    cwd: Path,
    compile_server_port: int,
    certfile: Path | None = None,
    keyfile: Path | None = None,
) -> Process:
    """Run the Flask server in a separate process."""
    process = Process(
        target=run_flask_in_thread,
        args=(port, cwd, compile_server_port, certfile, keyfile),
    )
    process.start()
    return process


def main() -> None:
    """Main function."""
    args = parse_args()
    run_flask_in_thread(args.port, args.fastled_js, args.certfile, args.keyfile)


if __name__ == "__main__":
    main()
